 
program aniso_modeling

  use aniso_modeling_mod
  use sep

  implicit none
  
  integer :: nx_pad,ny_pad,nz_pad_u,nz_pad_l,nx,ny,nz,nt,ns,ratio,nt_dat,nx_shot,ny_shot,nx_rec,ny_rec,nz_rec,zshot
  integer :: filter_length,npar,nref_start,niter_lloyd,want_map,want_filter,nref,want_attr
  integer,allocatable :: nstart(:),scales(:)
  real :: t,dt,sample_rate,dx,dy,dz,ox,oy,oz,dx_shot,dy_shot,ox_shot,oy_shot,dx_rec,dy_rec,dz_rec,ox_rec,oy_rec,oz_rec,shot_depth,damper
  real, allocatable :: tsou(:),dat(:,:,:,:,:,:),junk(:,:)

  call sep_init()

  call from_param("npar",npar)
  call from_param("nref_start",nref_start,20)
  call from_param("niter_lloyd",niter_lloyd,20)
  call from_param("want_map",want_map)
  call from_param("want_filter",want_filter)
  call from_param("want_attr",want_attr)

  call from_history("n1",nx)
  call from_history("n2",ny)
  call from_history("n3",nz)
  call from_history("d1",dx) !dispersion dx = vmin/fmax/3.5
  call from_history("d2",dy)
  call from_history("d3",dz)
  call from_history("o1",ox)
  call from_history("o2",oy)
  call from_history("o3",oz)

!  call from_param("nx_pad",nx_pad)
!  call from_param("ny_pad",ny_pad)
!  call from_param("nz_pad_u",nz_pad_u)
!  call from_param("nz_pad_l",nz_pad_l)
!  call from_param("damper",damper)

  call from_param("t",t)
  call from_param("dt",dt) !CFL dt*vmax/dx <= 0.45

  nt = t/dt+1

  call from_param("sample_rate",sample_rate)

  call init_space(ox,oy,oz,dx,dy,dz,nx,ny,nz)
  call init_time(dt,nt,sample_rate)
!  call init_taper(nx_pad,ny_pad,nz_pad_u,nz_pad_l,damper)

  call from_param("filter_length",filter_length) 

  allocate(junk(nx*ny*nz,npar))
  call sep_read(junk)

  allocate(scales(npar),nstart(npar))
  call from_param("nstart",nstart)
  call from_param("scales",scales)

  nref = lloyd_filter(junk,npar,nstart,nref_start,scales,niter_lloyd,want_map,filter_length) 
  write(0,*) 'after lloyd, nref', nref

  if (want_map == 1) then
    call sep_write(junk,"map")
    call to_history("n1",nx,"map")
    call to_history("n2",ny,"map")
    call to_history("n3",nz,"map")
    call to_history("n4",npar,"map")
    call to_history("o1",ox,"map")
    call to_history("o2",oy,"map")
    call to_history("o3",oz,"map")
    call to_history("o4",0.,"map")
    call to_history("d1",dx,"map")
    call to_history("d2",dy,"map")
    call to_history("d3",dz,"map")
    call to_history("d4",1.,"map")
  endif

  if (want_filter == 1) then
    call sep_write(filter_op,"filter")
    call to_history("n1",nref,"filter")
    call to_history("n2",(filter_length+1)/2,"filter")
    call to_history("n3",(filter_length+1)/2,"filter")
    call to_history("n4",(filter_length+1)/2,"filter")
    call to_history("o1",0.,"filter")
    call to_history("o2",0.,"filter")
    call to_history("o3",0.,"filter")
    call to_history("o4",0.,"filter")
    call to_history("d1",1.,"filter")
    call to_history("d2",1.,"filter")
    call to_history("d3",1.,"filter")
    call to_history("d4",1.,"filter")
  endif

  if (want_attr == 1) then
    call sep_write(attr,"attr")
    call to_history("n1",nref_start,"attr")
    call to_history("n2",npar,"attr")
    call to_history("o1",0.,"attr")
    call to_history("o2",0.,"attr")
    call to_history("d1",1.,"attr")
    call to_history("d2",1.,"attr")
  endif
  write(0,*) 'done lloyd'

  call from_param("ox_shot",ox_shot)
  call from_param("oy_shot",oy_shot)
  call from_param("dx_shot",dx_shot)
  call from_param("dy_shot",dy_shot)
  call from_param("nx_shot",nx_shot)
  call from_param("ny_shot",ny_shot)
  call from_param("shot_depth",shot_depth)
  zshot = (shot_depth-oz)/dz + 1.5 
  call init_shot(ox_shot,oy_shot,dx_shot,dy_shot,nx_shot,ny_shot,zshot)

  call from_param("ox_rec",ox_rec)
  call from_param("oy_rec",oy_rec)
  call from_param("oz_rec",oz_rec)
  call from_param("dx_rec",dx_rec)
  call from_param("dy_rec",dy_rec)
  call from_param("dz_rec",dz_rec)
  call from_param("nx_rec",nx_rec)
  call from_param("ny_rec",ny_rec)
  call from_param("nz_rec",nz_rec)
  call init_rec(ox_rec,oy_rec,oz_rec,dx_rec,dy_rec,dz_rec,nx_rec,ny_rec,nz_rec)

  ratio = nint(sample_rate/dt)
  nt_dat = nint(real(nt)/real(ratio)) + 1
  allocate(dat(nt_dat,nx_rec,ny_rec,nz_rec,nx_shot,ny_shot))

  call from_aux("source","n1",ns)
  allocate(tsou(ns))
  call sep_read(tsou,"source")
  call init_source(tsou,ns)

  call synthesize(dat)
  call to_history("n1",nt_dat)
  call to_history("n2",nx_rec)
  call to_history("n3",ny_rec)
  call to_history("n4",nz_rec)
  call to_history("o1",0)
  call to_history("o2",ox_rec)
  call to_history("o3",oy_rec)
  call to_history("o4",oz_rec)
  call to_history("d1",sample_rate)
  call to_history("d2",dx_rec)
  call to_history("d3",dy_rec)
  call to_history("d4",dz_rec)

  call sep_write(dat(:,:,:,:,1,1))


  deallocate(tsou,dat,junk,scales,nstart)

  call clean_up()

  call sep_close()

end program
