module dispersion_mod

  implicit none
  
  real, parameter,private :: pi = 4.*atan(1.),dk=0.002
  real :: c8(0:4), c10(0:5), c12(0:6), c14(0:7), c16(0:8),k(1001),a(1001)
  integer :: i,j

  contains 

  subroutine init_k()
    a = 0.
    do i=1,1001
        k(i) = -1. + (i-1.)*dk
        a(i) = -k(i)*k(i)
    enddo
  end subroutine

  subroutine init_c16()
    ! Calculate coefficients of 16th order finite difference
    	  
    c16(1) = 1.
    c16(2) = -7./40.
    c16(3) = 7./165.
    c16(4) = -7./704.
    c16(5) = 7./3575.
    c16(6) = -1./3432.
    c16(7) = 1./35035.
    c16(8) = -1./732160.
    c16(0) = -2*sum(c16(1:8))
    c16 = 16./9.*c16

    a = 0.
    do i=1,1001
        k(i) = -1. + (i-1.)*dk
        do j=1,8
          a(i) = a(i) + 2./pi/pi*c16(j)*(cos(k(i)*pi*j) - 1.)
        enddo
    enddo
	  
  end subroutine

  subroutine init_c14()
    ! Calculate coefficients of 14th order finite difference

    c14(1) = 1.
    c14(2) = -1./6.
    c14(3) = 1./27.
    c14(4) = -1./132.
    c14(5) = 1./825.
    c14(6) = -1./7722.
    c14(7) = -1./147147.
    c14(0) = -2*sum(c14(1:7))
    c14 = 7./4.*c14

    a = 0.
    do i=1,1001
        k(i) = -1. + (i-1.)*dk
        do j=1,7
          a(i) = a(i) + 2./pi/pi*c14(j)*(cos(k(i)*pi*j) - 1.)
        enddo
    enddo
	  
  end subroutine

  subroutine init_c12()
    ! Calculate coefficients of 12th order finite difference

    c12(1) = 1.
    c12(2) = -5./32.
    c12(3) = 5./162.
    c12(4) = -1./192.
    c12(5) = 1./1650.
    c12(6) = -1./28512.
    c12(0) = -2*sum(c12(1:6))
    c12 = 12./7.*c12

    a = 0.
    do i=1,1001
        k(i) = -1. + (i-1.)*dk
        do j=1,6
          a(i) = a(i) + 2./pi/pi*c12(j)*(cos(k(i)*pi*j) - 1.)
        enddo
    enddo
	  
  end subroutine

  subroutine init_c10()
    ! Calculate coefficients of 10th order finite difference

    c10(1) = 1.
    c10(2) = -1./7.
    c10(3) = 1./42.
    c10(4) = -1./336.
    c10(5) = 1./5250.
    c10(0) = -2*sum(c10(1:5))
    c10 = 5./3.*c10

    a = 0.
    do i=1,1001
        k(i) = -1. + (i-1.)*dk
        do j=1,5
          a(i) = a(i) + 2./pi/pi*c10(j)*(cos(k(i)*pi*j) - 1.)
        enddo
    enddo
	  
  end subroutine

  subroutine init_c8()
    ! Calculate coefficients of 8th order finite difference

    c8(1) = 1.
    c8(2) = -1./8.
    c8(3) = 1./63.
    c8(4) = -1./896.
    c8(0) = -2*sum(c8(1:4))
    c8 = 8./5.*c8

    a = 0.
    do i=1,1001
        k(i) = -1. + (i-1.)*dk
        do j=1,4
          a(i) = a(i) + 2./pi/pi*c8(j)*(cos(k(i)*pi*j) - 1.)
        enddo
    enddo
	  
  end subroutine

end module
