module evector_mod

  use base_fftwv3_mod
  use ieee_arithmetic
  use ieee_exceptions
  use cubic_mod

  implicit none

  real, private :: dx,dy,dz,ox,oy,oz,g(6)
  real, parameter :: pi = 4.*atan(1.)
  integer, private :: nx,ny,nz
  real, allocatable :: ufilter(:,:,:,:),filter(:,:,:,:)

  contains

  subroutine init_space(ox_in,oy_in,oz_in,dx_in,dy_in,dz_in,nx_in,ny_in,nz_in)

    real, intent(in) :: dx_in,dy_in,dz_in,ox_in,oy_in,oz_in
    integer, intent(in) :: nx_in,ny_in,nz_in
	
    dx = dx_in
    dy = dy_in
    dz = dz_in

    nx = nx_in
    ny = ny_in
    nz = nz_in

    ox = ox_in
    oy = oy_in
    oz = oz_in

  end subroutine  

  subroutine init_vel(eps1,eps2,del1,del2,del3,vpz)
  
    real, intent(in) :: eps1,eps2,del1,del2,del3,vpz

    g(6) = vpz*vpz
    g(1) = g(6)*(1.+2.*eps2)
    g(2) = g(1)*sqrt(1.+2.*del3)
    g(3) = g(6)*sqrt(1.+2.*del2)
    g(4) = g(6)*(1.+2.*eps1)
    g(5) = g(6)*sqrt(1.+2.*del1)

  end subroutine 

  subroutine init_evector(length,mode)

    integer, intent(in) :: length,mode
    real :: dkx,dky,dkz,kx2,ky2,kz2,t1,t2,p,s,detg,det1,det2,det3,a,b,c,x,err,kx,ky,kz,taperx,tapery,taperz,lambda,tempdet,d
    complex :: xc(2)
    complex, allocatable :: cfilter_k(:,:,:)
    integer :: ix,iy,iz,i,temp(1),half
    integer *8 :: plan

    half = (length-1)/2
    allocate(cfilter_k(length-1,length-1,length-1),filter(length-1,length-1,length-1,3))

    dkx = pi/dx/half
    dky = pi/dy/half
    dkz = pi/dz/half

    detg = g(1)*g(4)*g(6) + 2.*g(2)*g(3)*g(5) - g(1)*g(5)*g(5) - g(4)*g(3)*g(3) - g(6)*g(2)*g(2)
    det1 = g(4)*g(6) - g(5)*g(5)   
    det2 = g(1)*g(6) - g(3)*g(3)   
    det3 = g(4)*g(1) - g(2)*g(2)   

    filter = 0.

    do iz=1,half+1 
      kz = (iz-1.)*dkz
      kz2 = kz*kz
      do iy=1,half+1
        ky = (iy-1.)*dky
        ky2 = ky*ky
        do ix=1,half+1  
          kx = (ix-1.)*dkx
          kx2 = kx*kx
          a = g(1)*kx2 + g(4)*ky2 + g(6)*kz2
          b = kx2*ky2*det3 + ky2*kz2*det1 + kz2*kx2*det2
          c = detg*kx2*ky2*kz2
          call cubic(a,b,c,x,xc)
          d = (kx2 + ky2)*kz2
!          write(0,*) kx,ky,kz
!          write(0,*) x,xc
          if (mode == 1) lambda = real(xc(2))      
          if (mode == 2) lambda = real(xc(1))    
          tempdet = (lambda + g(1)*kx2)*(lambda + g(1)*ky2) - g(2)*kx2*g(2)*ky2
          if (tempdet == 0.) then
            if (kz == 0.) then
              filter(ix,iy,iz,:) = 0.    
            else
              filter(ix,iy,iz,:) = 1.  
            endif
            if (ky == 0.) then
              filter(ix,iy,iz,1) = 0.
              filter(ix,iy,iz,2) = 1.
            else
              filter(ix,iy,iz,1) = 1.
              filter(ix,iy,iz,2) = -(lambda + g(1)*kx2)/g(2)/ky2              
            endif    
          else   
            filter(ix,iy,iz,3) = 1    
            filter(ix,iy,iz,1) = kz2*(g(2)*g(5)*ky2 - g(3)*(lambda+g(4)*ky2))/tempdet    
            filter(ix,iy,iz,2) = kz2*(g(2)*g(3)*kx2 - g(5)*(lambda+g(1)*kx2))/tempdet                  
          endif       
        enddo
      enddo
    enddo
 
   do iz=1,half+1
     do iy=1,half+1
      do ix=half+2,length-1
        filter(ix,iy,iz,:) = filter(length+1-ix,iy,iz,:)
      enddo
     enddo
     do iy=half+2,length-1
      do ix=1,half+1
        filter(ix,iy,iz,:) = filter(ix,length+1-iy,iz,:)
      enddo
     enddo
     do iy=half+2,length-1
      do ix=half+2,length-1
        filter(ix,iy,iz,:) = filter(length+1-ix,length+1-iy,iz,:)
      enddo
     enddo
    enddo

    do iz=half+2,length-1
     do iy=1,length-1
      do ix=1,length-1
        filter(ix,iy,iz,:) = filter(ix,iy,length+1-iz,:)
      enddo
     enddo
    enddo

    call sfftw_dft_3dx_init(plan, cfilter_k, length-1, length-1, length-1, -1)  

    allocate(ufilter(length,length,length,3))

  if (mode == 1) then
  open(1,file="component1P.txt",form="formatted")
  open(2,file="component2P.txt",form="formatted")
  open(3,file="component3P.txt",form="formatted")
  else
  open(1,file="component1S.txt",form="formatted")
  open(2,file="component2S.txt",form="formatted")
  open(3,file="component3S.txt",form="formatted")
  endif

    do i=1,3

    cfilter_k = cmplx(filter(:,:,:,i))
    call sfftw_dft_3dx_oper(plan, cfilter_k, length-1, length-1, length-1)
    filter(:,:,:,i) = real(cfilter_k)/sqrt(real(length-1)**3)

    ufilter(half+1:length,half+1:length,half+1:length,i) = filter(1:half+1,1:half+1,1:half+1,i)

    do iz=half+1,length
     do iy=half+1,length
      do ix=1,half
        ufilter(ix,iy,iz,i) = ufilter(length+1-ix,iy,iz,i)
      enddo
     enddo
     do iy=1,half
      do ix=half+1,length
        ufilter(ix,iy,iz,i) = ufilter(ix,length+1-iy,iz,i)
      enddo
     enddo
     do iy=1,half
      do ix=1,half
        ufilter(ix,iy,iz,i) = ufilter(length+1-ix,length+1-iy,iz,i)
      enddo
     enddo
    enddo

    do iz=1,half
     do iy=1,length
      do ix=1,length
        ufilter(ix,iy,iz,i) = ufilter(ix,iy,length+1-iz,i)
      enddo
     enddo
    enddo    

    do iz=1,length
     do iy=1,length
      do ix=1,length
        write(i,*) ufilter(ix,iy,iz,i)
      enddo
     enddo
    enddo  

    enddo

  close(1)
  close(2)
  close(3)

    deallocate(cfilter_k,filter,ufilter)

  end subroutine 

end module
