module iso_mod

  use base_fftwv3_mod
  use ieee_arithmetic
  use ieee_exceptions

  implicit none

  real, private :: dx,dy,dz,dt,ox,oy,oz,dx_shot,dy_shot,ox_shot,oy_shot,dx_rec,dy_rec,dz_rec,ox_rec,oy_rec,oz_rec,c(0:8)
  real, parameter :: pi = 4.*atan(1.)
  integer, private :: nx,ny,nz,nt,nx_pad,ny_pad,nz_pad_u,nz_pad_l,nx_shot,ny_shot,nx_rec,ny_rec,nz_rec,zshot,ratio,half
  real, allocatable, private :: xtaper(:),ytaper(:),ztaper_u(:),ztaper_l(:),sou(:),k2(:,:,:)
  integer, allocatable, private :: shot_loc(:,:),rec_loc(:,:)

  contains

  subroutine init_c16()
    c = 0.
    c(1) = 1.
    c(2) = -7./40.
    c(3) = 7./165.
    c(4) = -7./704.
    c(5) = 7./3575.
    c(6) = -1./3432.
    c(7) = 1./35035.
    c(8) = -1./732160.
    c(0) = -2*sum(c(1:8))
    c = 16./9.*c	  
  end subroutine

  subroutine init_c14()
    c = 0.
    c(1) = 1.
    c(2) = -1./6.
    c(3) = 1./27.
    c(4) = -1./132.
    c(5) = 1./825.
    c(6) = -1./7722.
    c(7) = -1./147147.
    c(0) = -2*sum(c(1:7))
    c = 7./4.*c	  
  end subroutine

  subroutine init_c12()
    c = 0.
    c(1) = 1.
    c(2) = -5./32.
    c(3) = 5./162.
    c(4) = -1./192.
    c(5) = 1./1650.
    c(6) = -1./28512.
    c(0) = -2*sum(c(1:6))
    c = 12./7.*c
  end subroutine

  subroutine init_c10()
    c = 0.
    c(1) = 1.
    c(2) = -1./7.
    c(3) = 1./42.
    c(4) = -1./336.
    c(5) = 1./5250.
    c(0) = -2*sum(c(1:5))
    c = 5./3.*c
  end subroutine

  subroutine init_c8()
    c = 0.
    c(1) = 1.
    c(2) = -1./8.
    c(3) = 1./63.
    c(4) = -1./896.
    c(0) = -2*sum(c(1:4))
    c = 8./5.*c	  
  end subroutine

  subroutine init_c6()
    c = 0.
    c(1) = 1.
    c(2) = -1./10.
    c(3) = 1./135.
    c(0) = -2*sum(c(1:3))
    c = 3./2.*c
  end subroutine

  subroutine init_c4()
    c = 0.
    c(1) = 1.
    c(2) = -1./16.
    c(0) = -2*sum(c(1:2))
    c = 4./3.*c
  end subroutine

  subroutine init_c2()
    c = 0.
    c(1) = 1.
    c(0) = -2  
  end subroutine

  subroutine init_k2()
    integer :: ix,iy,iz
    real :: kx,ky,kz,halfx,halfy,halfz,dkx,dky,dkz
    halfx = (nx)/2 
    halfy = (ny)/2 
    halfz = (nz)/2 
    allocate(k2(nx,ny,nz))
    dkx = pi/dx/halfx
    dky = pi/dy/halfy
    dkz = pi/dz/halfz

    !$omp parallel do default(shared) private(ix,iy,iz,kx,ky,kz)
    do iz=1,halfz+1 
      kz = (iz-1.)*dkz
      do iy=1,halfy+1
        ky = (iy-1.)*dky
        do ix=1,halfx+1  
          kx = (ix-1.)*dkx
          k2(ix,iy,iz) = kx*kx+ky*ky+kz*kz
        enddo
      enddo
    enddo
    !$omp end parallel do

    !$omp parallel do default(shared) private(ix,iy,iz)
   do iz=1,halfz+1
     do iy=1,halfy+1
      do ix=halfx+2,nx
        k2(ix,iy,iz) = k2(nx+2-ix,iy,iz)
      enddo
     enddo
     do iy=halfy+2,ny
      do ix=1,halfx+1
        k2(ix,iy,iz) = k2(ix,ny+2-iy,iz)
      enddo
     enddo
     do iy=halfy+2,ny
      do ix=halfx+2,nx
        k2(ix,iy,iz) = k2(nx+2-ix,ny+2-iy,iz)
      enddo
     enddo
    enddo
    !$omp end parallel do

    !$omp parallel do default(shared) private(ix,iy,iz)
    do iz=halfz+2,nz
     do iy=1,ny
      do ix=1,nx
        k2(ix,iy,iz) = k2(ix,iy,nz+2-iz)
      enddo
     enddo
    enddo
    !$omp end parallel do
  end subroutine

  subroutine init_space(ox_in,oy_in,oz_in,dx_in,dy_in,dz_in,nx_in,ny_in,nz_in)
    real, intent(in) :: dx_in,dy_in,dz_in,ox_in,oy_in,oz_in
    integer, intent(in) :: nx_in,ny_in,nz_in	
    dx = dx_in
    dy = dy_in
    dz = dz_in

    nx = nx_in
    ny = ny_in
    nz = nz_in

    ox = ox_in
    oy = oy_in
    oz = oz_in
  end subroutine  

  subroutine init_time(dt_in,nt_in,rate_in)
    real, intent(in) :: dt_in,rate_in
    integer, intent(in) :: nt_in
    dt = dt_in
    nt = nt_in
    ratio = nint(rate_in/dt)
  end subroutine

  subroutine init_shot(ox_shot_in,oy_shot_in,dx_shot_in,dy_shot_in,nx_shot_in,ny_shot_in,zshot_in)
    real, intent(in) :: dx_shot_in,dy_shot_in,ox_shot_in,oy_shot_in
    integer, intent(in) :: nx_shot_in,ny_shot_in,zshot_in
    integer :: ix_shot,iy_shot
    dx_shot = dx_shot_in
    dy_shot = dy_shot_in
    nx_shot = nx_shot_in
    ny_shot = ny_shot_in
    ox_shot = ox_shot_in
    oy_shot = oy_shot_in
    zshot = zshot_in
    allocate(shot_loc(nx_shot*ny_shot,2))  
    do iy_shot=1,ny_shot
      do ix_shot=1,nx_shot
        shot_loc((iy_shot-1)*nx_shot+ix_shot,1) = (ox_shot+(ix_shot-1)*dx_shot - ox)/dx + 1.5 
        shot_loc((iy_shot-1)*nx_shot+ix_shot,2) = (oy_shot+(iy_shot-1)*dy_shot - oy)/dy + 1.5 
      enddo
    enddo 
  end subroutine 

  subroutine init_rec(ox_rec_in,oy_rec_in,oz_rec_in,dx_rec_in,dy_rec_in,dz_rec_in,nx_rec_in,ny_rec_in,nz_rec_in)  
    real, intent(in) :: dx_rec_in,dy_rec_in,dz_rec_in,ox_rec_in,oy_rec_in,oz_rec_in
    integer, intent(in) :: nx_rec_in,ny_rec_in,nz_rec_in
    integer :: ix_rec,iy_rec,iz_rec	
    dx_rec = dx_rec_in
    dy_rec = dy_rec_in
    dz_rec = dz_rec_in
    nx_rec = nx_rec_in
    ny_rec = ny_rec_in
    nz_rec = nz_rec_in
    ox_rec = ox_rec_in
    oy_rec = oy_rec_in
    oz_rec = oz_rec_in
    allocate(rec_loc(nx_rec*ny_rec*nz_rec,3))  
    do iz_rec=1,nz_rec
     do iy_rec=1,ny_rec
      do ix_rec=1,nx_rec
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,1) = (ox_rec+(ix_rec-1)*dx_rec - ox)/dx + 1.5 
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,2) = (oy_rec+(iy_rec-1)*dy_rec - oy)/dy + 1.5 
        rec_loc((iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec,3) = (oz_rec+(iz_rec-1)*dz_rec - oz)/dz + 1.5 
      enddo
     enddo
    enddo  
  end subroutine 

  subroutine init_source(tsou,ns)
    real, intent(in) :: tsou(:)
    integer, intent(in) :: ns
    allocate(sou(nt))
    sou = 0.
    sou(1:ns) = tsou
  end subroutine 

  subroutine init_taper(nx_pad_in,ny_pad_in,nz_pad_u_in,nz_pad_l_in,damper_in)
    real, intent(in) :: damper_in
    integer, intent(in) :: nx_pad_in,ny_pad_in,nz_pad_u_in,nz_pad_l_in
    real :: a0
    integer :: i
    nx_pad = nx_pad_in
    ny_pad = ny_pad_in
    nz_pad_u = nz_pad_u_in
    nz_pad_l = nz_pad_l_in 
    allocate(ztaper_u(nz_pad_u),ztaper_l(nz_pad_l),xtaper(nx_pad),ytaper(ny_pad))
    a0 = damper_in/real(nz_pad_u)
    do i=1,nz_pad_u
      ztaper_u(i) = exp(-(a0*real(nz_pad_u+1-i))**2)
    enddo
    a0 = damper_in/real(nz_pad_l)
    do i=1,nz_pad_l
      ztaper_l(i) = exp(-(a0*real(nz_pad_l+1-i))**2)
    enddo
    a0 = damper_in/real(nx_pad)
    do i=1,nx_pad
      xtaper(i) = exp(-(a0*real(nx_pad+1-i))**2)
    enddo
    a0 = damper_in/real(ny_pad)
    do i=1,ny_pad
      ytaper(i) = exp(-(a0*real(ny_pad+1-i))**2)
    enddo
  end subroutine 

  subroutine taper(field)
    real, intent(inout) :: field(nx,ny,nz)
    integer :: ix,iy,iz
    !$omp parallel do default(shared) private(ix,iy)
    do iy=1,ny
      do ix=1,nx 
        field(ix,iy,1:nz_pad_u) = field(ix,iy,1:nz_pad_u)*ztaper_u
        field(ix,iy,nz-nz_pad_l+1:nz) = field(ix,iy,nz-nz_pad_l+1:nz)*ztaper_l(nz_pad_l:1:-1)
      enddo
    enddo
    !$omp end parallel do
    !$omp parallel do default(shared) private(iz,iy)
    do iz=nz_pad_u+1,nz-nz_pad_l
      do iy=1,ny
        field(1:nx_pad,iy,iz) = field(1:nx_pad,iy,iz)*xtaper
        field(nx-nx_pad+1:nx,iy,iz) = field(nx-nx_pad+1:nx,iy,iz)*xtaper(nx_pad:1:-1)
      enddo
    enddo
    !$omp end parallel do
    !$omp parallel do default(shared) private(iz,ix)
    do iz=nz_pad_u+1,nz-nz_pad_l
      do ix=nx_pad+1,nx-nx_pad
        field(ix,1:ny_pad,iz) = field(ix,1:ny_pad,iz)*ytaper
        field(ix,ny-ny_pad+1:ny,iz) = field(ix,ny-ny_pad+1:ny,iz)*ytaper(ny_pad:1:-1)
      enddo
    enddo
    !$omp end parallel do
  end subroutine 

  subroutine clean_up()
    deallocate(shot_loc,rec_loc,sou)
    if (allocated(xtaper)) deallocate(xtaper,ytaper,ztaper_u,ztaper_l)
    if (allocated(k2)) deallocate(k2)
  end subroutine 

  subroutine synthesize_x(v,dat)

    real, intent(inout) :: dat(:,:,:,:,:,:),v
    real :: v2dt2
    integer:: ix_shot,iy_shot,it,ishot,ix_rec,iy_rec,iz_rec,irec
    real, allocatable :: cur(:,:,:),prev(:,:,:),next(:,:,:),temp(:,:,:)

    allocate(cur(nx,ny,nz), prev(nx,ny,nz), next(nx,ny,nz), temp(nx,ny,nz))

    cur = 0.
    prev = 0.
    next = 0.
    temp = 0.
 
    v2dt2 = v*v*dt*dt
    do iy_shot=1,ny_shot
     do ix_shot=1,nx_shot
      ishot=(iy_shot-1)*nx_shot+ix_shot
      do it=1,nt
        write(0,*) it

        cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot) = cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot) + sou(it) 

        call laplacian_x(cur,temp)
        next = v2dt2*temp + 2.*cur - prev

        if (maxval(abs(next)) > 1.e17) then
          write(0,*) "blow up"
          write(0,*) maxval(abs(next))
          exit  ! Exit the loop when the amplitude blows up
        endif

        if (.not. ieee_is_finite(sum(next))) then
          write(0,*) "nan"
          exit ! Exit the loop when NaN is found
        endif

!        call taper(cur)
!        call taper(next)

        if (mod(it-1,ratio) == 0) then
          do iz_rec=1,nz_rec
           do iy_rec=1,ny_rec
            do ix_rec=1,nx_rec
              irec=(iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec
              dat((it-1)/ratio+1,ix_rec,iy_rec,iz_rec,ix_shot,iy_shot)  = cur(rec_loc(irec,1),rec_loc(irec,2),rec_loc(irec,3))
            enddo
           enddo
          enddo
        endif

        prev = cur ! Update wavefield
        cur = next

      enddo ! end time loop

     enddo ! end shot loop
    enddo
   
    deallocate(prev,cur,next,temp)

  end subroutine 

  subroutine init_stencil(order)
    integer,intent(in) :: order
    half = order/2
    select case (order)
      case(16) 
        call init_c16()
      case(14) 
        call init_c14()
      case(12) 
        call init_c12()
      case(10) 
        call init_c10()
      case(8) 
        call init_c8()
      case(6) 
        call init_c6()
      case(4) 
        call init_c4()
      case(2) 
        call init_c2()
      case(0) 
        call init_k2()
    end select
  end subroutine 

  subroutine laplacian_x(p,dp) !FD spatial laplacian 

    real, intent(in) :: p(nx,ny,nz)
    real, intent(inout) :: dp(nx,ny,nz)
    integer :: ix,iy,iz,ic
    real :: dx_2,dy_2,dz_2

    dx_2 = 1./dx/dx
    dy_2 = 1./dy/dy
    dz_2 = 1./dz/dz

    !$omp parallel do default(shared) private(ix,iy,iz,ic)
    do iz=half+1,nz-half 
      do iy=half+1,ny-half
        do ix=half+1,nx-half
          dp(ix,iy,iz) = c(0)*p(ix,iy,iz)*(dx_2+dy_2+dz_2)
          do ic=1,half
            dp(ix,iy,iz) = dp(ix,iy,iz) + c(ic)*((p(ix+ic,iy,iz)+p(ix-ic,iy,iz))*dx_2+&
                                                 (p(ix,iy+ic,iz)+p(ix,iy-ic,iz))*dy_2+&
                                                 (p(ix,iy,iz+ic)+p(ix,iy,iz-ic))*dz_2)
          enddo
        enddo
      enddo
    enddo
    !$omp end parallel do

  end subroutine 

  subroutine synthesize_k(v,dat)

    real, intent(inout) :: dat(:,:,:,:,:,:),v
    integer:: ix_shot,iy_shot,it,ishot,ix_rec,iy_rec,iz_rec,irec
    real, allocatable :: cur(:,:,:),prev(:,:,:),next(:,:,:)
    complex, allocatable :: cfield(:,:,:)
    integer *8 :: forward,inverse
    real :: v2dt2

    allocate(cur(nx,ny,nz), prev(nx,ny,nz), next(nx,ny,nz), cfield(nx,ny,nz))

    cur = 0.
    prev = 0.
    next = 0.

    call sfftw_dft_3dx_init(forward, cfield, nx, ny, nz, 1)  
    call sfftw_dft_3dx_init(inverse, cfield, nx, ny, nz, -1)  

    v2dt2 = v*v*dt*dt
    do iy_shot=1,ny_shot
     do ix_shot=1,nx_shot
      ishot=(iy_shot-1)*nx_shot+ix_shot
      do it=1,nt
        write(0,*) it

        cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot) = cur(shot_loc(ishot,1),shot_loc(ishot,2),zshot) + sou(it) 

        cfield = cmplx(cur)
        call sfftw_dft_3dx_oper(forward, cfield, nx, ny, nz)
        cfield = cmplx(real(cfield)*(-k2))
        call sfftw_dft_3dx_oper(inverse, cfield, nx, ny, nz)

        next = real(cfield)*v2dt2 + 2.*cur - prev

        if (maxval(abs(next)) > 1.e17) then
          write(0,*) "blow up"
          write(0,*) maxval(abs(next))
          exit  ! Exit the loop when the amplitude blows up
        endif

        if (.not. ieee_is_finite(sum(next))) then
          write(0,*) "nan"
          exit ! Exit the loop when NaN is found
        endif

!        call taper(cur)
 !       call taper(next)

        if (mod(it-1,ratio) == 0) then
          do iz_rec=1,nz_rec
           do iy_rec=1,ny_rec
            do ix_rec=1,nx_rec
              irec=(iz_rec-1)*nx_rec*ny_rec+(iy_rec-1)*nx_rec+ix_rec
              dat((it-1)/ratio+1,ix_rec,iy_rec,iz_rec,ix_shot,iy_shot)  = cur(rec_loc(irec,1),rec_loc(irec,2),rec_loc(irec,3))
            enddo
           enddo
          enddo
        endif

        prev = cur ! Update wavefield
        cur = next

      enddo ! end time loop

     enddo ! end shot loop
    enddo
   
    deallocate(prev,cur,next,cfield)

  end subroutine 

end module
