 
program shear_reduction

  use iso_mod
  use sep

  implicit none
  
  integer :: nx_pad,ny_pad,nz_pad_u,nz_pad_l,nx,ny,nz,nt,ns,ratio,nt_dat,nx_shot,ny_shot,nx_rec,ny_rec,nz_rec,zshot,order
  real :: t,dt,sample_rate,dx,dy,dz,ox,oy,oz,dx_shot,dy_shot,ox_shot,oy_shot,dx_rec,dy_rec,dz_rec,ox_rec,oy_rec,oz_rec,shot_depth,damper,v
  real, allocatable :: tsou(:),dat(:,:,:,:,:,:),out(:,:,:,:)

  call sep_init()

  call from_param("order",order,8)

  call from_param("n1",nx)
  call from_param("n2",ny)
  call from_param("n3",nz)
  call from_param("d1",dx) !dispersion dx = vmin/fmax/3.5
  call from_param("d2",dy)
  call from_param("d3",dz)
  call from_param("o1",ox)
  call from_param("o2",oy)
  call from_param("o3",oz)

!  call from_param("nx_pad",nx_pad)
!  call from_param("ny_pad",ny_pad)
!  call from_param("nz_pad_u",nz_pad_u)
!  call from_param("nz_pad_l",nz_pad_l)
!  call from_param("damper",damper)

  call from_param("t",t)
  call from_param("dt",dt) !CFL dt*vmax/dx <= 0.45

  nt = t/dt+1

  call from_param("sample_rate",sample_rate)

  call init_space(ox,oy,oz,dx,dy,dz,nx,ny,nz)
  call init_time(dt,nt,sample_rate)
!  call init_taper(nx_pad,ny_pad,nz_pad_u,nz_pad_l,damper)

  call from_param("v",v)  

  call init_stencil(order)

  call from_param("ox_shot",ox_shot)
  call from_param("oy_shot",oy_shot)
  call from_param("dx_shot",dx_shot)
  call from_param("dy_shot",dy_shot)
  call from_param("nx_shot",nx_shot)
  call from_param("ny_shot",ny_shot)
  call from_param("shot_depth",shot_depth)
  zshot = (shot_depth-oz)/dz + 1.5 
  call init_shot(ox_shot,oy_shot,dx_shot,dy_shot,nx_shot,ny_shot,zshot)

  call from_param("ox_rec",ox_rec)
  call from_param("oy_rec",oy_rec)
  call from_param("oz_rec",oz_rec)
  call from_param("dx_rec",dx_rec)
  call from_param("dy_rec",dy_rec)
  call from_param("dz_rec",dz_rec)
  call from_param("nx_rec",nx_rec)
  call from_param("ny_rec",ny_rec)
  call from_param("nz_rec",nz_rec)
  call init_rec(ox_rec,oy_rec,oz_rec,dx_rec,dy_rec,dz_rec,nx_rec,ny_rec,nz_rec)

  ratio = nint(sample_rate/dt)
  nt_dat = nint(real(nt)/real(ratio)) + 1
  allocate(dat(nt_dat,nx_rec,ny_rec,nz_rec,nx_shot,ny_shot),out(nt_dat,nx_rec,ny_rec,nz_rec))

  call from_aux("source","n1",ns)
  allocate(tsou(ns))
  call sep_read(tsou,"source")
  call init_source(tsou,ns)

  if (order == 0) then
    call synthesize_k(v,dat)
  else 
    call synthesize_x(v,dat)
  endif

  out = dat(:,:,:,:,1,1)
  call to_history("n1",nt_dat)
  call to_history("n2",nx_rec)
  call to_history("n3",ny_rec)
  call to_history("n4",nz_rec)
!  call to_history("n5",nx_shot)
!  call to_history("n6",ny_shot)
  call to_history("o1",0)
  call to_history("o2",ox_rec)
  call to_history("o3",oy_rec)
  call to_history("o4",oz_rec)
!  call to_history("o5",ox_shot)
!  call to_history("o6",oy_shot)
  call to_history("d1",sample_rate)
  call to_history("d2",dx_rec)
  call to_history("d3",dy_rec)
  call to_history("d4",dz_rec)
!  call to_history("d5",dx_shot)
!  call to_history("d6",dy_shot)

  call sep_write(out)

  deallocate(tsou,dat,out)

  call clean_up()

  call sep_close()

end program
